/*
colormachinesdl.cc
Copyright 2007-2021 Matt Heinzen

This file is part of Color Machine.

Color Machine is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Color Machine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Color Machine; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "colormachinesdl.h"
#include "utils.h"
#include "palettes.h"
#include "modules.h"

#include <string>
#include <sstream>
#include <vector>
#include <iostream>

#include <time.h>
#include <stdio.h>


const int DEFAULT_WIDTH = 1920;
const int DEFAULT_HEIGHT = 1080;
const int MAX_WIDTH = 1920;
const int NUM_EFFECTS = 4;
const int MIN_MODULE_SECONDS = 20;
const int MAX_MODULE_SECONDS = 40;
const int MIN_CYCLES = 90;
const int MAX_CYCLES = 240;
const int FADE_IN_MILLIS = 500;
const int FADE_OUT_MILLIS = 2000;


SDL_Surface *gScreen;
SDL_Window *gWindow;
SDL_Renderer *gRenderer;
Uint8 *gScreenBuf;
SDL_Color gPalette[256];
int gSurfaceW, gSurfaceH;


int getRandomNewItem(std::vector<int> items, int previous = -1) {
	const int MAX_TRIES = 1000;

	if (items.size() == 1) return items[0];

	// if items contains duplicate entries, those entries will have a greater change of being
	// selected, but should still never be selected twice in a row *unless* items is filled with
	// multiple copies of only one value
	int item = previous;
	for (int i = 0; item == previous && i < MAX_TRIES; i++) item = items[randomRange(items.size())];

	return item;
}

void drawScreen(SDL_Color *palette)
{
	SDL_SetPaletteColors(gScreen->format->palette, palette, 0, 256);

	// stupidly inefficient to create a new texture every frame, but updating an existing texture was not working right
	SDL_Texture* texture = SDL_CreateTextureFromSurface(gRenderer, gScreen);
	SDL_RenderCopy(gRenderer, texture, NULL, NULL);
	SDL_RenderPresent(gRenderer);
	SDL_DestroyTexture(texture);
}

void frameSleep(Uint32 frameStartMillis, Uint32 deltaMillis)
{
	Uint32 nowMillis = SDL_GetTicks();
	Uint32 frameEndMillis = frameStartMillis + deltaMillis;
	if (frameEndMillis > nowMillis) SDL_Delay(frameEndMillis - nowMillis);
}

bool checkQuit()
{
	SDL_Event event;

	while (SDL_PollEvent(&event) == 1) {
		switch (event.type) {
		case SDL_KEYDOWN:
		case SDL_MOUSEBUTTONDOWN:
			printf("keyboard or mouse button\n");
			return true;
		case SDL_QUIT:
			printf("quit\n");
			return true;
		case SDL_MOUSEMOTION:
			break;
		}
	}

	return false;
}

void mainLoop(std::vector<int> modules, std::vector<int> palettes, int const moduleSecCmd, int const cyclesCmd, int const fpsCmd,
			  bool linearPalette)
{
	SDL_Color paletteCopy[256];

	// reset previous module and palette or set by parameter
	int moduleIdx = -1;
	int paletteIdx = -1;

	// flush the event queue
	SDL_Event event;
	while (SDL_PollEvent(&event) == 1);

	// keep looping until a keypress, mouse click or quit event
	while (true)
	{
		// choose a new module and palette
		moduleIdx = getRandomNewItem(modules, moduleIdx);
		paletteIdx = getRandomNewItem(palettes, paletteIdx);

		printf("module = %d, palette = %d\n", moduleIdx, paletteIdx);

		// clear the screen
		initPaletteBlack(gPalette);
		memset(gScreenBuf, 0, gSurfaceW * gSurfaceH);
		drawScreen(gPalette);

		// draw the chosen module and setup the initial unfaded palette
		drawModule(moduleIdx);
		initPalette(gPalette, paletteIdx, linearPalette);

		// generate semi-random lengths of time for the entire module and for individual frames
		int numModuleSec = moduleSecCmd ? moduleSecCmd : randomInt(MIN_MODULE_SECONDS, MAX_MODULE_SECONDS);
		int numFrameMillis = 1000 / (fpsCmd ? fpsCmd : randomInt(61, 121));

		// loop until the module time has elapsed
		Uint32 moduleStartMillis = SDL_GetTicks();
		Uint32 moduleEndMillis = moduleStartMillis + numModuleSec * 1000;
		while (SDL_GetTicks() < moduleEndMillis)
		{
			// select RGB component rotation directions and speeds
			int R, G, B;
			do {
				R = randomRange(5) - 2;
				G = randomRange(5) - 2;
				B = randomRange(5) - 2;
			} while (abs(R) + abs(G) + abs(B) < 2);

			// DEBUG: prevent animation to visualize unmodified palette
			//R = G = B = 0;

			// choose a random effect length
			int numCycles = cyclesCmd ? cyclesCmd : randomInt(MIN_CYCLES, MAX_CYCLES);

			for (int i = 0; i < numCycles; i++) {
				Uint32 frameStartMillis = SDL_GetTicks();

				// capture relevant events
				if (checkQuit()) return;
				else if (frameStartMillis > moduleEndMillis) i = numCycles;

				// update palette
				switch (R)
				{
				case -2: rotateRed(gPalette, true); // fallthrough (double rotate)
				case -1: rotateRed(gPalette, true); break;
				case 2: rotateRed(gPalette, false); // fallthrough (double rotate)
				case 1: rotateRed(gPalette, false); break;
				}
				switch (G)
				{
				case -2: rotateGreen(gPalette, true); // fallthrough (double rotate)
				case -1: rotateGreen(gPalette, true); break;
				case 2: rotateGreen(gPalette, false); // fallthrough (double rotate)
				case 1: rotateGreen(gPalette, false); break;
				}
				switch (B)
				{
				case -2: rotateBlue(gPalette, true); // fallthrough (double rotate)
				case -1: rotateBlue(gPalette, true); break;
				case 2: rotateBlue(gPalette, false); // fallthrough (double rotate)
				case 1: rotateBlue(gPalette, false); break;
				}

				// fade palette at beginning and end of module
				SDL_Color *palette = gPalette;
				bool isFadeIn = frameStartMillis < moduleStartMillis + FADE_IN_MILLIS;
				bool isFadeOut = frameStartMillis > moduleEndMillis - FADE_OUT_MILLIS;
				if (isFadeIn || isFadeOut)
				{
					copyPalette(gPalette, paletteCopy);
					fadePalette_x256(paletteCopy, isFadeIn
						? (FADE_IN_MILLIS - (frameStartMillis - moduleStartMillis)) * 256 / FADE_IN_MILLIS
						: (FADE_OUT_MILLIS - (moduleEndMillis - frameStartMillis)) * 256 / FADE_OUT_MILLIS);
					palette = paletteCopy;
				}

				drawScreen(palette);

				// calculate ease in/out
				// TODO: improve this, doesn't work well with vsync locked frame rates
				/**
				float effectPct = (float)i / numCycles;
				const float easeCyclesPct = 0.4;
				const float easeMaxMillis = 33;
				float easePct = effectPct < easeCyclesPct ? (easeCyclesPct - effectPct) / easeCyclesPct
					: effectPct > 1.0 - easeCyclesPct ? (effectPct - 1.0 + easeCyclesPct) / easeCyclesPct
					: 0.0;
				int adjNumFrameMillis = numFrameMillis + (int)(easePct * easeMaxMillis);
				/**/

				// no ease in/out
				int adjNumFrameMillis = numFrameMillis;

				frameSleep(frameStartMillis, adjNumFrameMillis);
			}
		}
	}
}

void handleNullError(void *p, char const *objName)
{
	if (p == NULL)
	{
		printf("Unable to create %s: %s\n", objName, SDL_GetError());
		exit(1);
	}
}

std::vector<int> splitInts(std::string const &s)
{
	std::vector<int> v;
	std::istringstream iss(s);
	std::string token;
	while (std::getline(iss, token, ',')) v.push_back(std::stoi(token));

	return v;
}

std::vector<int> createIntVec(int last) {
	std::vector<int> v;
	for (int i = 1; i <= last; i++) v.push_back(i);
	return v;
}

int main(int argc, char **argv)
{
	// TODO: use command line arguments for selecting all parameters (time as range, cycles as range, fps as range)

	// initialze SDL
	if (SDL_Init(SDL_INIT_VIDEO) != 0)
	{
		printf("Unable to initialize SDL video\n");
		exit(EXIT_FAILURE);
	}

	// make sure SDL_Quit gets called when the program exits
	atexit(SDL_Quit);

	SDL_DisplayMode displayMode;
	SDL_GetCurrentDisplayMode(0, &displayMode);
	printf("screen: %dx%d\n", displayMode.w, displayMode.h);

	// set default windowed mode size
	int windowW = std::min((int)(displayMode.w * 0.9f), MAX_WIDTH);
	int windowH = std::min((int)(displayMode.h * 0.9f), MAX_WIDTH * displayMode.h / displayMode.w);

	if (argc > 1 && strstr(argv[1], "--help") == argv[1])
	{
		if (strstr(argv[1], "--help-modules") == argv[1])
		{
			printf("The following modules can be selected with the 'm=...' option, like 'm=5' or 'm=1,2,10,23:'\n");
			for (int i = MODULE_NONE + 1; i < MODULE_MAX; i++) printf("  %d: %s\n", i, getModuleName(i));
		}
		else if (strstr(argv[1], "--help-palettes") == argv[1])
		{
			printf("The following palettes can be selected with the 'p=...' option, like 'p=2' or 'p=3,6,7:'\n");
			for (int i = PALETTE_NONE + 1; i < PALETTE_MAX; i++) printf("  %d: %s\n", i, getPaletteName(i));
		}
		else
		{
			printf(
				"Usage: colormachinesdl [options]\n"
				"Options\n"
				"  --help\t\tDisplay this help\n"
				"  --help-modules\tShow list of modules to use with m=\n"
				"  --help-palettes\tShow list of palettes to use with p=\n"
				"  w={0|1}\t\tWindow mode: 0=Fullscreen (default), 1=Windowed\n"
				"  s=W,H\t\t\tScreen size (Windowed mode only)\n"
				"  m=...\t\t\tComma-separated list of modules to use (default is all modules), see --help-modules\n"
				"  p=...\t\t\tComma-separated list of palettes to use (default is all palettes), see --help-palettes\n"
				"  l={0|1}\t\tPalette curve linearity: 0=Non-linear (default), 1=Linear\n"
				"  t=N\t\t\tModule Time in seconds (default is randomized between 20-40)\n"
				"  f=N\t\t\tFrames Per Second (NOT CURRENTLY WORKING)\n"
				"  c=N\t\t\tEffect Cycles (default is randomized between 90 and 240)\n"
				"  r=N\t\t\tRandom seed (defaults to timer)\n");
		}

		exit(EXIT_SUCCESS);
	}

	// initialize screensaver parameters from command line args
	int windowed, linear, moduleTime, cycles, fps;
	windowed = linear = moduleTime = cycles = fps = 0;
	int randSeed = -1;
	std::vector<int> modules, palettes;
	for (int i = 1; i < argc; i++)
	{
		char *pStr = argv[i];
		if (strstr(pStr, "w=") == pStr) sscanf(pStr + 2, "%d", &windowed);
		if (strstr(pStr, "s=") == pStr)
		{
			std::vector<int> winSize = splitInts(pStr + 2);
			if (winSize.size() != 2) printf("Must specify window size as s=WIDTH,HEIGHT! Using default values\n");
			else
			{
				windowW = winSize[0];
				windowH = winSize[1];
			}
		}
		if (strstr(pStr, "m=") == pStr) modules = splitInts(pStr + 2);
		if (strstr(pStr, "p=") == pStr) palettes = splitInts(pStr + 2);
		if (strstr(pStr, "l=") == pStr) sscanf(pStr + 2, "%d", &linear);
		if (strstr(pStr, "t=") == pStr) sscanf(pStr + 2, "%d", &moduleTime);
		if (strstr(pStr, "f=") == pStr) sscanf(pStr + 2, "%d", &fps);
		if (strstr(pStr, "c=") == pStr) sscanf(pStr + 2, "%d", &cycles);
		if (strstr(pStr, "r=") == pStr) sscanf(pStr + 2, "%d", &randSeed);
	}

	// TODO: error checking of params, especially FPS and any negative values

	// assign all modules and/or palettes by default
	if (modules.empty()) modules = createIntVec(26);
	if (palettes.empty()) palettes = createIntVec(8);

	gWindow = windowed
		? SDL_CreateWindow("Color Machine", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, windowW, windowH, 0)
		: SDL_CreateWindow("", 0, 0, 0, 0, SDL_WINDOW_FULLSCREEN_DESKTOP);
	handleNullError(gWindow, "window");

	gRenderer = SDL_CreateRenderer(gWindow, -1, SDL_RENDERER_PRESENTVSYNC); // SDL_RENDERER_PRESENTVSYNC has no effect? might be related to current drivers
	handleNullError(gRenderer, "renderer");

	//SDL_GL_SetSwapInterval(0); // set but no effect

	int winW, winH;
	SDL_GetWindowSize(gWindow, &winW, &winH);
	int screenW = winW < MAX_WIDTH ? winW : MAX_WIDTH;
	int screenH = screenW * winH / winW;
	printf("window: %dx%d, surface target: %dx%d\n", winW, winH, screenW, screenH);

	gScreen = SDL_CreateRGBSurface(0, screenW, screenH, 8, 0, 0, 0, 0);
	handleNullError(gScreen, "surface");

	gSurfaceW = gScreen->w;
	gSurfaceH = gScreen->h;

	gScreenBuf = (Uint8 *)gScreen->pixels;
	
	// hide the mouse pointer
	SDL_ShowCursor(SDL_DISABLE);

	printf("surface is %dx%d\n", gSurfaceW, gSurfaceH);

	// randomize with user selected seed or system clock
	srandom(randSeed > -1 ? randSeed : time(NULL));

	// start loop
	mainLoop(modules, palettes, moduleTime, cycles, fps, linear);

	return 0;
}
