/*
utils.cc
Copyright 2007-2021 Matt Heinzen

This file is part of Color Machine.

Color Machine is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Color Machine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Color Machine; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "utils.h"

#include "colormachinesdl.h"

#include <math.h>


int randomRange(const int range)
{
	// from http://www.thinkage.ca/english/gcos/expl/c/lib/rand.html
	// uses random() instead of rand()

	int val;

	do val = random() / (RAND_MAX / range); while (val >= range);
	return val;
}

int randomInt(const int low, const int high)
{
	return low + randomRange(high - low + 1);
}

float randomFloat(const float low, const float high)
{
	return low + (float)random() / RAND_MAX * (high - low);
}

bool randomBool()
{
	return random() % 2 == 1;
}

int randomSign()
{
	return randomBool() ? 1 : -1;
}

Shape randomShape()
{
	return (Shape)randomRange(SHAPE_MAX);
}

Uint8 randomColor() {
	return (Uint8)randomInt(1, 255);
}

void cycleColor(Uint8 &colorIndex, int const amount)
{
	colorIndex = (Uint8)(((int)(colorIndex - 1) + amount + 255) % 255 + 1);
}

Uint8 int2Color(int a)
{
	if (a < 0) a += (-a / 255 + 1) * 255;
	return a % 255 + 1;
}

void plotPixel(int const x, int const y, Uint8 const colorIndex)
{
	if (x >= 0 && x < gSurfaceW && y >= 0 && y < gSurfaceH)
		gScreenBuf[y * gSurfaceW + x] = colorIndex;
}

void drawHLine(int x1, int x2, int y, Uint8 const colorIndex)
{
	if (y < 0 || y >= gSurfaceH) return;

	if (x1 > x2)
	{
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}

	if (x1 >= gSurfaceW || x2 < 0) return;

	if (x1 < 0) x1 = 0;
	if (x2 >= gSurfaceW) x2 = gSurfaceW - 1;

	memset(gScreenBuf + y * gSurfaceW + x1, colorIndex, x2 - x1 + 1);
}

void drawVLine(int X, int y1, int y2, Uint8 const colorIndex)
{
	if (X < 0 || X >= gSurfaceW) return;

	if (y1 > y2) {
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}

	if (y1 >= gSurfaceH || y2 < 0) return;

	if (y1 < 0) y1 = 0;
	if (y2 >= gSurfaceH) y2 = gSurfaceH - 1;

	for (int i = y1; i <= y2; i++)
		gScreenBuf[i * gSurfaceW + X] = colorIndex;
}

void drawLine(int x1, int y1, int x2, int y2, Uint8 const colorIndex)
{
	// TODO: use a decent algorithm

	double slope;

	if (x1 == x2) drawVLine(x1, y1, y2, colorIndex);
	if (y1 == y2) drawHLine(x1, x2, y1, colorIndex);

	if (abs(x2 - x1) >= abs(y2 - y1))
	{
		if (x1 > x2)
		{
			int temp = x1;
			x1 = x2;
			x2 = temp;
			temp = y1;
			y1 = y2;
			y2 = temp;
		}

		slope = (double)(y2 - y1) / (double)(x2 - x1);
		double Y = y1;

		for (int px = x1; px <= x2; px++, Y += slope)
			plotPixel(px, (int)Y, colorIndex);
	}
	else
	{
		if (y1 > y2)
		{
			int temp = x1;
			x1 = x2;
			x2 = temp;
			temp = y1;
			y1 = y2;
			y2 = temp;
		}

		slope = (double)(x2 - x1) / (double)(y2 - y1);
		double X = x1;

		for (int py = y1; py <= y2; py++, X += slope)
			plotPixel((int)X, py, colorIndex);
	}
}

void drawBox(int x1, int y1, int x2, int y2, Uint8 const colorIndex)
{
	if (x1 > x2)
	{
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}

	if (y1 > y2)
	{
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}

	if (x1 >= gSurfaceW || x2 < 0 || y1 >= gSurfaceH || y2 < 0)
		return;

	if (x1 < 0) x1 = 0;
	if (y1 < 0) y1 = 0;
	if (x2 >= gSurfaceW) x2 = gSurfaceW - 1;
	if (y2 >= gSurfaceH) y2 = gSurfaceH - 1;

	for (int i = y1; i <= y2; i++)
		memset(gScreenBuf + i * gSurfaceW + x1, colorIndex, x2 - x1 + 1);
}

void drawDisc(int x, int y, int r, Uint8 colorIndex)
{
	if (r <= 0) return;

	int minY = y - r < 0 ? 0 : y - r;
	int maxY = y + r >= gSurfaceH ? gSurfaceH - 1 : y + r;

	for (int i = minY; i <= maxY; i++)
	{
		int dy = y - i;
		double halfWidth = sqrt(r * r - dy * dy);
		drawHLine((int)(x - halfWidth), (int)(x + halfWidth), i, colorIndex);
	}
}

void drawDiamond(int x, int y, int r, Uint8 colorIndex)
{
	for (int i = 0; i <= r; i++)
	{
		drawHLine(x - i, x + i, y - r + i, colorIndex);
		drawHLine(x - i, x + i, y + r - i, colorIndex);
	}
}

void drawHexagon(int x, int y, int r, Uint8 colorIndex)
{
	for (int i = 0; i <= r; i++)
	{
		int halfWidth = (i + r) * 0.55f;
		drawHLine(x - halfWidth, x + halfWidth, y - r + i, colorIndex);
		drawHLine(x - halfWidth, x + halfWidth, y + r - i, colorIndex);
	}
}

void drawOctagon(int x, int y, int r, Uint8 colorIndex)
{
	for (int i = 0; i <= r / 2; i++)
	{
		int halfWidth = r / 2 + i;
		drawHLine(x - halfWidth, x + halfWidth, y - r + i, colorIndex);
		drawHLine(x - halfWidth, x + halfWidth, y + r - i, colorIndex);
	}
	drawBox(x - r, y - r / 2, x + r, y + r / 2, colorIndex);
}

void drawShape(Shape s, int x, int y, int r, Uint8 colorIndex)
{
	switch (s)
	{
	case SHAPE_CIRCLE: drawDisc(x, y, r, colorIndex); break;
	case SHAPE_SQUARE: drawBox(x - r, y - r, x + r, y + r, colorIndex); break;
	case SHAPE_DIAMOND: drawDiamond(x, y, r, colorIndex); break;
	case SHAPE_HEXAGON: drawHexagon(x, y, r, colorIndex); break;
	case SHAPE_OCTAGON: drawOctagon(x, y, r, colorIndex); break;
	}
}
